/*
 * Copyright (C) 2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.acidhouse;

import java.util.ConcurrentModificationException;
import java.util.List;

/**
 * {@code DatastoreSession} is central client-interfaced object to control 
 * Google App Engine Datastore through Acid House and represents interactive 
 * session between client and Google App Engine Datastore.
 * Basic CRUD operations are like this: 
 * <pre>
 * import org.eiichiro.acidhouse.DatastoreSession;
 * import org.eiichiro.acidhouse.appengine.AppEngineDatastoreSession;
 * ...
 * 
 * // Create entity.
 * Entity1 entity11 = new Entity1();
 * entity11.id = "Key11";
 * entity11.i = 11;
 * Entity1 entity12 = new Entity1();
 * entity12.id = "Key12";
 * entity12.i = 12;
 * DatastoreSession session = new AppEngineDatastoreSession();
 * session.begin();
 * 
 * try {
 * 	session.put(entity11);
 * 	session.put(entity12);
 * 	session.commit();
 * } catch (Exception e) {
 * 	session.rollback();
 * 	throw e;
 * } finally {
 * 	session.close();
 * }
 * 
 * 
 * // Read entity.
 * DatastoreSession session = new AppEngineDatastoreSession();
 * Entity1 entity11 = session.get(Entity1.class, "Key11");
 * Entity1 entity12 = session.get(Entity1.class, "Key12");
 * session.close();
 * 
 * 
 * // Update entity.
 * DatastoreSession session = new AppEngineDatastoreSession();
 * session.begin();
 * 
 * try {
 * 	Entity1 entity11 = session.get(Entity1.class, "Key11");
 * 	Entity1 entity12 = session.get(Entity1.class, "Key12");
 * 	entity11.i = 1111;
 * 	entity12.i = 1222;
 * 	session.update(entity11);
 * 	session.update(entity12);
 * 	session.commit();
 * } catch (Exception e) {
 * 	session.rollback();
 * 	throw e;
 * } finally {
 * 	session.close();
 * }
 * 
 * 
 * // Delete entity.
 * DatastoreSession session = new AppEngineDatastoreSession();
 * session.begin();
 * 
 * try {
 * 	Entity1 entity11 = session.get(Entity1.class, "Key11");
 * 	Entity1 entity12 = session.get(Entity1.class, "Key12");
 * 	session.delete(entity11);
 * 	session.delete(entity12);
 * 	session.commit();
 * } catch (Exception e) {
 * 	session.rollback();
 * 	throw e;
 * } finally {
 * 	session.close();
 * }
 * </pre>
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public interface DatastoreSession {

	/**
	 * Begins transaction.
	 * You have to do transactional operation after calling this method.
	 */
	public void begin();
	
	/**
	 * Commits current transaction.
	 * You have to invoke this method after calling {@code #begin()} method.
	 * 
	 * @throws IndoubtException When data consistency has broken in current 
	 * transaction.
	 */
	public void commit() throws IndoubtException;
	
	/**
	 * Rolls back current transaction.
	 * You have to invoke this method after calling {@code #begin()} method.
	 */
	public void rollback();
	
	/**
	 * Returns entity instance corresponding to the specified key.
	 * It depends on the implementation specification of this interface whether 
	 * an exception is thrown or {@code null} is returned if the entity is not 
	 * found.
	 * 
	 * @param <E> The entity type.
	 * @param clazz The entity type that you want to get.
	 * @param id The id corresponding to the entity that you want to get.
	 * @return The entity instance corresponding to the specified key.
	 * @throws ConcurrentModificationException If the entity corresponding to 
	 * the specified key is being modified by the other transaction.
	 */
	public <E> E get(Class<E> clazz, Object id) throws ConcurrentModificationException;
	
	/**
	 * Puts the specified entity instance into Datastore newly.
	 * This method must be invoked under a transaction. Depending on the 
	 * implementation of this interface, if the entity that has the same key as 
	 * the specified entity has been already stored, this method might throw 
	 * {@code EntityExistsException}.
	 * 
	 * @param entity The entity instance to be put into Datastore.
	 * @throws EntityExistsException If the entity that has the same key as the 
	 * specified entity has already existed.
	 */
	public void put(Object entity) throws EntityExistsException;
	
	/**
	 * Updates entity with the specified entity instance.
	 * This method must be invoked under a transaction. 
	 * 
	 * @param entity The entity instance to be updated.
	 */
	public void update(Object entity);
	
	/**
	 * Deletes the specified entity from Datastore.
	 * This method must be invoked under a transaction.
	 * 
	 * @param entity The entity to be deleted.
	 */
	public void delete(Object entity);
	
	/** Closes this session. */
	public void close();
	
	/**
	 * Returns {@code GetList} for the specified entity class.
	 * This method is the entry point for getting-list Command Builder API.
	 * 
	 * @param <E> The entity type to get with this {@code GetList}.
	 * @param clazz The entity type to get with this {@code GetList}. 
	 * @return {@code GetList} for the specified entity class.
	 */
	public <E> GetList<E> get(Class<E> clazz);
	
	/**
	 * Returns {@code GetScalar} for this specified {@code Aggregation}.
	 * This method is the entry point for aggregation Command Builder API.
	 * 
	 * @param <R> The property value type to aggregate with this 
	 * {@code GetScalar}.
	 * @param aggregation The {@code Aggregation} this method executes.
	 * @return {@code GetScalar} for the specified {@code Aggregation} 
	 * instance.
	 */
	public <E, R> GetScalar<E, R> get(Aggregation<R> aggregation);
	
	/**
	 * Returns {@code Update} for the specified entity class.
	 * This method is the entry point for updating Command Builder API.
	 * 
	 * @param <E> The type of entity updated by this {@code Update}.
	 * @param clazz The type of entity updated by this {@code Update}.
	 * @return The {@code Update} for the specified entity class.
	 */
	public <E> Update<E> update(Class<E> clazz);
	
	/**
	 * Returns {@code Delete} for the specified entity class.
	 * This method is the entry point for deleting Command Builder API.
	 * 
	 * @param <E> The type of entity deleted by this {@code Delete}.
	 * @param clazz The type of entity deleted by this {@code Delete}.
	 * @return The {@code Delete} for the specified entity class.
	 */
	public <E> Delete<E> delete(Class<E> clazz);
	
	/**
	 * Puts the specified entity instances into Datastore newly.
	 * This method must be invoked under the transactional session. Depending on 
	 * the implementation of this interface, the entity that has the same key as 
	 * the specified entity has been already stored, this method might throw any 
	 * exception.
	 * 
	 * @param <E> The type of entity instances to be put by this operation.
	 * @param entities The entity instances that you want to put into Datastore.
	 * @throws EntityExistsException If the entity that has the same key as the 
	 * specified entity has been already existed.
	 */
	public <E> void put(E... entities) throws EntityExistsException;
	
	/**
	 * Puts the specified entity list into Datastore newly.
	 * This method must be invoked under the transactional session. Depending on 
	 * the implementation of this interface, the entity that has the same key as 
	 * the specified entity has been already stored, this method might throw any 
	 * exception.
	 * 
	 * @param <E> The type of entity instances to be put by this operation.
	 * @param entities The entity instances that you want to put into 
	 * Datastore as a {@code List} view.
	 * @throws EntityExistsException If the entity that has the same key as the 
	 * specified entity has been already existed.
	 */
	public <E> void put(List<E> entities) throws EntityExistsException;
	
	/**
	 * Updates entities with the specified entity instances.
	 * This method must be invoked under the transaction.
	 * 
	 * @param <E> Type of updated entity.
	 * @param entities The entities to be updated.
	 */
	public <E> void update(E... entities);
	
	/**
	 * Updates entities states with the specified entity instance list.
	 * This method must be invoked under the transaction.
	 * 
	 * @param <E> Type of updated entity.
	 * @param entities The entity list to be updated.
	 */
	public <E> void update(List<E> entities);
	
	/**
	 * Deletes the specified entities from Datastore. 
	 * This method must be invoked under the transaction.
	 * 
	 * @param <E> Type of deleted entity. 
	 * @param entities The entities to be deleted.
	 */
	public <E> void delete(E... entities);
	
	/**
	 * Deletes the specified entity list from Datastore. 
	 * This method must be invoked under the transaction.
	 * 
	 * @param <E> Type of deleted entity. 
	 * @param entities The entity list to be deleted.
	 */
	public <E> void delete(List<E> entities);
	
	/**
	 * 
	 * @return
	 */
	public boolean isTransactional();
	
}
