/*
 * Copyright (C) 2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.acidhouse;

import java.util.ArrayList;
import java.util.List;

/**
 * {@code IndoubtException} is thrown by {@code DatastoreSession} when the data 
 * consistency is broken (The situation that there are committed operation and 
 * uncommitted operation in one transaction). If you get this exception, you 
 * have to recover your data consistency immediately from the transaction dump.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class IndoubtException extends RuntimeException {

	private static final long serialVersionUID = 8694562251382720993L;
	
	private final List<Operation> uncommitted;

	/**
	 * Constructs a new {@code IndoubtException} instance with the specified 
	 * cause, transaction and committed count.
	 * 
	 * @param exception The cause exception.
	 * @param transaction The current datastore transaction.
	 * @param commit The committed count.
	 */
	public IndoubtException(Exception exception, DatastoreTransaction transaction, int commit) {
		super(dump(transaction, commit), exception);
		List<Operation> operations = transaction.getOperations();
		List<Operation> uncommitted = new ArrayList<Operation>();
		
		for (int i = commit - 1; i < operations.size(); i++) {
			uncommitted.add(operations.get(i));
		}
		
		this.uncommitted = uncommitted;
	}
	
	private static String dump(DatastoreTransaction transaction, int commit) {
		StringBuilder builder = new StringBuilder();
		List<Operation> operations = transaction.getOperations();
		builder.append("Transaction [" + transaction.getId() + "] is in-doubt; Dump: \n");
		
		for (int i = 0; i < operations.size(); i++) {
			Operation operation = operations.get(i);
			builder.append("\tOperation " + (i + 1) + " ["
					+ operation.getType() + "] -> ["
					+ ((i < commit) ? "COMMITTED" : "UNCOMMITTED")
					+ "]: Entity [" + Entities.getId(operation.getEntity())
					+ "]\n");
		}
		
		return builder.toString();
	}

	/**
	 * Returns uncommitted operations as list.
	 * 
	 * @return The uncommitted operations as list.
	 */
	public List<Operation> getUncommitted() {
		return uncommitted;
	}
	
}
