/*
 * Copyright (C) 2009-2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.acidhouse.appengine;

import org.eiichiro.acidhouse.ComparableFilter;
import org.eiichiro.acidhouse.Filter;
import org.eiichiro.acidhouse.Order;
import org.eiichiro.acidhouse.ComparableFilter.Comparison;

/**
 * {@code QueryRestriction} indicates whether the specified {@code Filter} and 
 * {@code Order} can be applied to Google App Engine Datastore query or not.
 * If the specified {@code Filter} is restricted to Google App Engine Datastore 
 * query, the {@code Filter} is processed as in-memory filtering. So is 
 * {@code Order}.See the restriction detail 
 * <a href="http://code.google.com/intl/en/appengine/docs/java/datastore/queriesandindexes.html#Restrictions_on_Queries">
 * here</a>.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class QueryRestriction {

	private Filter<?> filter;
	
	private Order<?, ?> order;
	
	private boolean restricted = false;
	
	/**
	 * Indicates whether the specified {@code Filter} is restricted to Google 
	 * App Engine Datastore query or not.
	 * This method indicates the following cases as <b>restricted</b>: 
	 * <ul>
	 * <li>
	 * Property that the specified filter targets is not root entity's property.
	 * </li>
	 * <li>
	 * Comparison type of the specified filter is other than 
	 * {@code Comparison#EQUAL_TO} and the property that the specified filter 
	 * targets is not the same as the previous filter's one.
	 * </li>
	 * </ul>
	 * 
	 * @param filter The filter to be indicated whether restricted or not.
	 * @return Whether the specified {@code Filter} is restricted to Google App 
	 * Engine Datastore query or not.
	 */
	public boolean restricted(Filter<?> filter) {
		if (!filter.getProperty().getMetamodel().isRoot()) {
			return true;
		}
		
		if (filter instanceof ComparableFilter<?>) {
			ComparableFilter<?> comparableFilter = (ComparableFilter<?>) filter;
			Comparison comparison = comparableFilter.getComparison();
			
			if (comparison != Comparison.EQUAL_TO) {
				if (this.filter == null) {
					this.filter = filter;
				} else {
					if (this.filter.getProperty() != filter.getProperty()) {
						return true;
					}
				}
			}
		}
		
		return false;
	}
	
	/**
	 * Indicates whether the specified {@code Order} is restricted to Google App 
	 * Engine Datastore query or not.
	 * This method indicates the following cases as <b>restricted</b>: 
	 * <ul>
	 * <li>
	 * Property that the specified order targets is not root entity's property.
	 * </li>
	 * <li>
	 * If any filter has been already specified, the property that the filter 
	 * targets is not the same as the property that the order targets.
	 * </li>
	 * <li>
	 * All orders after the order indicated as restricted once.
	 * </li>
	 * </ul>
	 * 
	 * @param order The order to be indicated whether restricted or not.
	 * @return Whether the specified {@code Order} is restricted to Google App 
	 * Engine Datastore query or not.
	 */
	public boolean restricted(Order<?, ?> order) {
		if (!order.getProperty().getMetamodel().isRoot()) {
			restricted = true;
			return true;
		}
		
		if (!restricted) {
			if (this.order == null) {
				if (filter != null) {
					if (filter.getProperty() != order.getProperty()) {
						restricted = true;
						return true;
					}
				}
				
				this.order = order;
			}
			
			return false;
		}
		
		return true;
	}
	
}
