/*
 * Copyright (C) 2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.acidhouse.appengine;

import java.lang.reflect.Field;
import java.lang.reflect.ParameterizedType;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

import org.eiichiro.acidhouse.ComparableFilter;
import org.eiichiro.acidhouse.DatastoreTransaction;
import org.eiichiro.acidhouse.Entities;
import org.eiichiro.acidhouse.Id;
import org.eiichiro.acidhouse.Lock;
import org.eiichiro.acidhouse.Operation;
import org.eiichiro.acidhouse.Order;
import org.eiichiro.acidhouse.Referential;
import org.eiichiro.acidhouse.Transient;
import org.eiichiro.reverb.lang.UncheckedException;

import com.google.appengine.api.datastore.Blob;
import com.google.appengine.api.datastore.Entity;
import com.google.appengine.api.datastore.EntityTranslator;
import com.google.appengine.api.datastore.Key;
import com.google.appengine.api.datastore.Query;
import com.google.appengine.api.datastore.Query.FilterOperator;
import com.google.appengine.api.datastore.Query.SortDirection;
import com.google.storage.onestore.v3.OnestoreEntity;
import com.google.storage.onestore.v3.OnestoreEntity.EntityProto;

/**
 * {@code Translation} has translation utilities between Acid House data types 
 * and Google App Engine Datastore data types.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public abstract class Translation {

	/** Google App Engine Datastore kind for {@code Lock}. */
	public static final String LOCK_KIND = "_" + Lock.class.getSimpleName();
	
	/** Google App Engine Datastore entity's property name for {@code Lock#transaction}. */
	public static final String TRANSACTION_PROPERTY = "transaction";
	
	/** Google App Engine Datastore entity's property name for {@code Lock#timestamp}. */
	public static final String TIMESTAMP_PROPERTY = "timestamp";
	
	/** Google App Engine Datastore kind for {@code DatastoreTransaction}. */
	public static final String DATASTORE_TRANSACTION_KIND = "_" + DatastoreTransaction.class.getSimpleName();
	
	/** Google App Engine Datastore kind for {@code Operation}. */
	public static final String OPERATION_KIND = "_" + Operation.class.getSimpleName();
	
	/** Google App Engine Datastore entity's property name for {@code Operation#type}. */
	public static final String TYPE_PROPERTY = "type";
	
	/** Google App Engine Datastore entity's property name for {@code Lock#entity}'s class. */
	public static final String CLASS_PROPERTY = "class";
	
	/** Google App Engine Datastore entity's property name for {@code Lock#entity}'s Protocol Buffers message. */
	public static final String PROTO_PROPERTY = "proto";
	
	private Translation() {}
	
	/**
	 * Translates Google App Engine Datastore entities to Acid House entity with 
	 * {@code AppEngineDatastoreService}.
	 * 
	 * @param <E> The type of Acid House entity.
	 * @param clazz The {@code Class} of Acid House entity.
	 * @param entities Google App Engine Datastore entities.
	 * @param datastore {@code AppEngineDatastoreService}.
	 * @return Acid House entity translated from Google App Engine Datastore 
	 * entities.
	 */
	public static <E> E toObject(Class<E> clazz, List<Entity> entities,
			AppEngineDatastoreService datastore) {
		return toObject(clazz, entities, datastore, 0);
	}
	
	@SuppressWarnings("unchecked")
	private static <E> E toObject(Class<E> clazz, List<Entity> entities,
			AppEngineDatastoreService datastore, int generation) {
		if (entities.size() == 0) {
			return null;
		}
		
		try {
			Iterator<Entity> iterator = entities.iterator();
			Entity entity = iterator.next();
			String kind = entity.getKind();
			
			if (kind.equals(LOCK_KIND) || kind.equals(DATASTORE_TRANSACTION_KIND)) {
				return null;
			}
			
			E object = clazz.newInstance();
			List<Entity> descendant = new ArrayList<Entity>();
			
			while (iterator.hasNext()) {
				descendant.add(iterator.next());
			}
			
			for (Field field : clazz.getDeclaredFields()) {
				field.setAccessible(true);
				Class<?> type = field.getType();
				
				if (field.getAnnotation(Id.class) != null) {
					String name = entity.getKey().getName();
					field.set(object, (name == null) ? entity.getKey().getId() : name);
					continue;
				}
				
				if (field.getAnnotation(Transient.class) != null) {
					continue;
				}
				
				if (field.getAnnotation(Referential.class) != null) {
					Key key = (Key) entity.getProperty(field.getName());
					field.set(object, toObject(type, datastore.query(new Query(key)), datastore));
					continue;
				}
				
				if (Arrays.binarySearch(CORE_VALUE_TYPES, type.getName()) >= 0) {
					Object value = entity.getProperty(field.getName());
					
					// int and java.lang.Integer fields are translated as 
					// int64Value (long) property in EntityProto, so I have to 
					// revert them to int...
					if ((type.getName().equals(CORE_VALUE_TYPES[16]) 
							|| type.getName().equals(CORE_VALUE_TYPES[20])) 
							&& (value.getClass().getName().equals(CORE_VALUE_TYPES[21]) 
									|| value.getClass().getName().equals(CORE_VALUE_TYPES[25]))) {
						long l = (Long) value;
						value = (int) l;
					}
					
					field.set(object, value);
				} else if (Arrays.binarySearch(CORE_COLLECTION_TYPES, type.getName()) >= 0) {
					if (entity.hasProperty(field.getName())) {
						field.set(object, entity.getProperty(field.getName()));
						continue;
					}
					
					Class<?> collectionType = null;
					
					if (type.equals(List.class)) {
						collectionType = ArrayList.class;
					} else if (type.equals(Set.class)) {
						collectionType = HashSet.class;
					} else if (type.equals(SortedSet.class)) {
						collectionType = TreeSet.class;
					} else {
						collectionType = type;
					}
					
					Collection<Object> collection = (Collection<Object>) collectionType.newInstance();
					Class<?> generic = (Class<?>) ((ParameterizedType) 
							field.getGenericType()).getActualTypeArguments()[0];
					List<List<Entity>> children = new ArrayList<List<Entity>>();
					List<Entity> child = null;
					Key previous = null;
					
					for (Entity e : descendant) {
						Key key = Keys.getPedigree(e).get(generation + 1);
						
						if (!key.getKind().equals(toKind(generic))) {
							// Entities are assumed to be sorted by key.
							break;
						}
						
						if (!key.equals(previous)) {
							child = new ArrayList<Entity>();
							children.add(child);
						}
						
						child.add(e);
					}
					
					for (List<Entity> list : children) {
						collection.add(toObject(generic, list, datastore, generation + 1));
					}
					
					field.set(object, collection);
				} else if (type.getName().startsWith(ARRAY_TYPE)) {
					throw new IllegalArgumentException(
							"Array type translation [" + type
									+ "] is not supported");
				} else {
					List<Entity> child = new ArrayList<Entity>();
					
					for (Entity e : descendant) {
						Key key = Keys.getPedigree(e).get(generation + 1);
						
						if (!key.getKind().equals(toKind(type))) {
							// Entities are assumed to be sorted by key.
							break;
						}
						
						child.add(e);
					}
					
					field.set(object, toObject(type, child, datastore, generation + 1));
				}
			}
			
			return object;
		} catch (Exception e) {
			throw new UncheckedException(e);
		}
	}
	
	/**
	 * Translates Acid House entity to Google App Engine Datastore entities.
	 * 
	 * @param entity Acid House entity.
	 * @return Google App Engine Datastore entities translated from Acid House 
	 * entity.
	 */
	public static List<Entity> toEntities(Object entity) {
		return toEntities(null, entity);
	}
	
	/** Types available for Google App Engine Datastore. */
	public static String[] CORE_VALUE_TYPES = {
		"boolean", 
		"com.google.appengine.api.datastore.Blob", 
		"com.google.appengine.api.datastore.Category", 
		"com.google.appengine.api.datastore.Email", 
		"com.google.appengine.api.datastore.GeoPt", 
		"com.google.appengine.api.datastore.IMHandle", 
		"com.google.appengine.api.datastore.Key", 
		"com.google.appengine.api.datastore.Link", 
		"com.google.appengine.api.datastore.PhoneNumber", 
		"com.google.appengine.api.datastore.PostalAddress", 
		"com.google.appengine.api.datastore.Rating", 
		"com.google.appengine.api.datastore.ShortBlob", 
		"com.google.appengine.api.datastore.Text", 
		"com.google.appengine.api.users.User", 
		"double", 
		"float", 
		"int", 
		"java.lang.Boolean", 
		"java.lang.Double", 
		"java.lang.Float", 
		"java.lang.Integer", 
		"java.lang.Long", 
		"java.lang.Short", 
		"java.lang.String", 
		"java.util.Date", 
		"long", 
		"short"
	};
	
	/** Collection types available for Google App Engine Datastore. */
	public static String[] CORE_COLLECTION_TYPES = {
		"java.util.ArrayList", 
		"java.util.HashSet", 
		"java.util.LinkedHashSet", 
		"java.util.LinkedList", 
		"java.util.List", 
		"java.util.Set", 
		"java.util.SortedSet", 
		"java.util.Stack", 
		"java.util.TreeSet", 
		"java.util.Vector"
	};
	
	/** Prefix of array type. */
	public static String ARRAY_TYPE = "[";
	
	/**
	 * Translates Acid House entity to Google App Engine Datastore entities with 
	 * the specified {@code Key} of parent.
	 * 
	 * @param parent The {@code Key} of parent.
	 * @param entity Acid House entity.
	 * @return Google App Engine Datastore entities translated from Acid House 
	 * entity.
	 */
	public static List<Entity> toEntities(Key parent, Object entity) {
		List<Entity> entities = new ArrayList<Entity>();
		Class<?> clazz = entity.getClass();
		
		if (clazz.getAnnotation(org.eiichiro.acidhouse.Entity.class) == null) {
			throw new IllegalArgumentException("Entity [" + clazz
					+ "] must be annotated @org.eiichiro.acidhouse.Entity");
		}
		
		Entity e = new Entity(Keys.create(parent, toKind(clazz), Entities.getId(entity)));
		entities.add(e);
		
		for (Field field : clazz.getDeclaredFields()) {
			if (field.getAnnotation(Id.class) != null) {
				continue;
			}
			
			if (field.getAnnotation(Transient.class) != null) {
				continue;
			}
			
			field.setAccessible(true);
			Class<?> type = field.getType();
			
			try {
				if (field.getAnnotation(Referential.class) != null) {
					Key reference = Keys.create(toKind(type), Entities.getId(field.get(entity)));
					e.setProperty(field.getName(), reference);
					continue;
				}
				
				if (Arrays.binarySearch(CORE_VALUE_TYPES, type.getName()) >= 0) {
					e.setProperty(field.getName(), field.get(entity));
				} else if (Arrays.binarySearch(CORE_COLLECTION_TYPES, type.getName()) >= 0) {
					Collection<?> children = (Collection<?>) field.get(entity);
					
					if (Arrays.binarySearch(CORE_VALUE_TYPES, 
							((Class<?>) ((ParameterizedType) field.getGenericType())
									.getActualTypeArguments()[0]).getName()) >= 0) {
						e.setProperty(field.getName(), children);
					} else {
						for (Object child : children) {
							entities.addAll(toEntities(e.getKey(), child));
						}
					}
					
				} else if (type.getName().startsWith(ARRAY_TYPE)) {
					throw new IllegalArgumentException(
							"Array type translation [" + type
									+ "] is not supported");
				} else {
					entities.addAll(toEntities(e.getKey(), field.get(entity)));
				}

			} catch (Exception exception) {
				throw new UncheckedException(exception);
			}
		}
		
		return entities;
	}
	
	/**
	 * Translates Google App Engine Datastore entity to Acid House {@code Lock} 
	 * entity.
	 * 
	 * @param entity oogle App Engine Datastore entity.
	 * @return Acid House {@code Lock} entity.
	 */
	public static Lock toLock(Entity entity) {
		return new Lock(entity.getKey().getName(), 
				(String) entity.getProperty(TRANSACTION_PROPERTY), 
				(Date) entity.getProperty(TIMESTAMP_PROPERTY));
	}
	
	/**
	 * Translates Acid House {@code Lock} entity to Google App Engine Datastore 
	 * entity with the specified {@code Key} of parent.
	 * 
	 * @param lock Acid House {@code Lock} entity.
	 * @param parent The {@code Key} of parent.
	 * @return Google App Engine Datastore entity translated from Acid House 
	 * {@code Lock} entity.
	 */
	public static Entity toEntity(Lock lock, Key parent) {
		Entity entity = new Entity(LOCK_KIND, lock.getId(), parent);
		entity.setProperty(TRANSACTION_PROPERTY, lock.getTransaction());
		entity.setProperty(TIMESTAMP_PROPERTY, lock.getTimestamp());
		return entity;
	}
	
	/**
	 * Translates Google App Engine Datastore entities to Acid House 
	 * {@code DatastoreTransaction} entity with {@code AppEngineDatastoreService}.
	 * 
	 * @param transactions Google App Engine Datastore entities.
	 * @param datastore {@code AppEngineDatastoreService}.
	 * @return Acid House {@code DatastoreTransaction} entity translated from 
	 * Google App Engine Datastore entities.
	 */
	@SuppressWarnings("unchecked")
	public static DatastoreTransaction toDatastoreTransaction(
			Iterable<Entity> transactions, AppEngineDatastoreService datastore) {
		DatastoreTransaction datastoreTransaction = null;
		Map<Long, Entity> operations = new TreeMap<Long, Entity>();
		
		for (Entity transaction : transactions) {
			if (transaction.getKind().equals(DATASTORE_TRANSACTION_KIND)) {
				datastoreTransaction = new DatastoreTransaction(transaction.getKey().getName());
			} else if (transaction.getKind().equals(OPERATION_KIND)) {
				operations.put(transaction.getKey().getId(), transaction);
			}
		}
		
		for (Long sequence : operations.keySet()) {
			Entity operation = operations.get(sequence);
			List<Entity> entity = new ArrayList<Entity>();
			
			for (Blob proto : (List<Blob>) operation.getProperty(PROTO_PROPERTY)) {
				OnestoreEntity.EntityProto entityProto = new OnestoreEntity.EntityProto();
				entityProto.mergeFrom(proto.getBytes());
				entity.add(EntityTranslator.createFromPb(entityProto));
			}
			
			try {
				datastoreTransaction.add(new Operation(sequence, Enum.valueOf(
						Operation.Type.class, (String) operation.getProperty(TYPE_PROPERTY)),
						toObject(Class.forName((String) operation.getProperty(CLASS_PROPERTY)), 
								entity, datastore)));
			} catch (ClassNotFoundException e) {
				throw new UncheckedException(e);
			}
		}
		
		return datastoreTransaction;
	}
	
	/**
	 * Translates Acid House {@code DatastoreTransaction} entity to Google App 
	 * Engine Datastore entities with the specified {@code Key} of parent.
	 * 
	 * @param transaction Acid House {@code Lock} entity.
	 * @param parent The {@code Key} of parent.
	 * @return Google App Engine Datastore entities translated from Acid House 
	 * {@code DatastoreTransaction} entity.
	 */
	public static List<Entity> toEntities(DatastoreTransaction transaction, Key parent) {
		List<Entity> entities = new ArrayList<Entity>();
		Key key = Keys.create(parent, DATASTORE_TRANSACTION_KIND, transaction.getId());
		entities.add(new Entity(key));
		
		for (Operation operation : transaction.getOperations()) {
			Entity entity = new Entity(Keys.create(key, OPERATION_KIND, operation.getSequence()));
			entity.setProperty(TYPE_PROPERTY, operation.getType().name());
			entity.setProperty(CLASS_PROPERTY, operation.getEntity().getClass().getName());
			List<Blob> blobs = new ArrayList<Blob>();
			entity.setProperty(PROTO_PROPERTY, blobs);
			entities.add(entity);
			
			for (Entity e : toEntities(operation.getEntity())) {
				EntityProto proto = EntityTranslator.convertToPb(e);
				byte[] bytes = proto.toByteArray();
				blobs.add(new Blob(bytes));
//				System.out.println("EntityTranslator.convertToPb(e).toByteArray(): [" 
//						+ bytes.length + "] bytes");
//				System.out.println("Entity [" + entity.getKey() + "]: [" 
//						+ EntityTranslator.convertToPb(entity).toByteArray().length 
//						+ "] bytes");
			}
		}
		
		return entities;
	}
	
	/**
	 * Translates {@code Class} to Google App Engine Datastore entity's kind 
	 * name.
	 * 
	 * @param clazz The {@code Class} of Acid House entity.
	 * @return The kind name translated from Acid House entity class.
	 */
	public static String toKind(Class<?> clazz) {
		if (clazz.getAnnotation(org.eiichiro.acidhouse.Entity.class) == null) {
			throw new IllegalArgumentException("Entity class [" + clazz
					+ "] must be annotated by @org.eiichiro.acidhouse.Entity");
		}
		
		return clazz.getSimpleName();
	}
	
	/**
	 * Translates {@code ComparableFilter.Comparison} to Google App Engine 
	 * Datastore {@code FilterOperator}.
	 * 
	 * @param comparison Acid House {@code ComparableFilter.Comparison}.
	 * @return Google App Engine Datastore {@code FilterOperator} translated 
	 * from Acid House {@code ComparableFilter.Comparison}.
	 */
	public static FilterOperator toFilterOperator(ComparableFilter.Comparison comparison) {
		if (comparison == ComparableFilter.Comparison.EQUAL_TO) {
			return FilterOperator.EQUAL;
		} else if (comparison == ComparableFilter.Comparison.GREATER_THAN) {
			return FilterOperator.GREATER_THAN;
		} else if (comparison == ComparableFilter.Comparison.GREATER_THAN_OR_EQUAL_TO) {
			return FilterOperator.GREATER_THAN_OR_EQUAL;
		} else if (comparison == ComparableFilter.Comparison.LESS_THAN) {
			return FilterOperator.LESS_THAN;
		} else if (comparison == ComparableFilter.Comparison.LESS_THAN_OR_EQUAL_TO) {
			return FilterOperator.LESS_THAN_OR_EQUAL;
		} else if (comparison == ComparableFilter.Comparison.NOT_EQUAL_TO) {
			return FilterOperator.NOT_EQUAL;
		} else {
			throw new UnsupportedOperationException("Comparison [" + comparison
					+ "] is not supported by Google App Engine Datastore");
		}
	}
	
	/**
	 * Translates {@code Order.Direction} to Google App Engine Datastore 
	 * {@code SortDirection}.
	 * 
	 * @param direction Acid House {@code Order.Direction}.
	 * @return Google App Engine Datastore {@code SortDirection} translated 
	 * from Acid House {@code Order.Direction}.
	 */
	public static SortDirection toSortDirection(Order.Direction direction) {
		if (direction == Order.Direction.ASC) {
			return SortDirection.ASCENDING;
		} else if (direction == Order.Direction.DESC) {
			return SortDirection.DESCENDING;
		} else {
			throw new UnsupportedOperationException("Direction [" + direction
					+ "] is not supported by Google App Engine Datastore");
		}
	}
	
}
