/*
 * Copyright (C) 2009-2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.acidhouse.metamodel;

/**
 * {@code Metamodel} represents a metamodel of Datastore entity.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public abstract class Metamodel<E> {

	private final Class<E> kind;
	
	private final Metamodel<?> parent;
	
	private final String name;
	
	/**
	 * Constructs a new {@code Metamodel} instance with the specified entity 
	 * kind. This constructor is called when the represented entity is root 
	 * entity (not owned child entity property).
	 * 
	 * @param kind The entity kind.
	 */
	public Metamodel(Class<E> kind) {
		this(kind, null, "");
	}
	
	/**
	 * Constructs a new {@code Metamodel} instance with the specified entity 
	 * kind, parent metamodel and metamodel name. This constructor is called 
	 * when the represented entity is declared as owned child entity property.
	 * 
	 * @param kind The entity kind.
	 * @param parent The parent entity metamodel when this instance is declared 
	 * as owned child entity property.
	 * @param name The metamodel name when this instance is declared as owned 
	 * child entity property.
	 */
	public Metamodel(Class<E> kind, Metamodel<?> parent, String name) {
		this.kind = kind;
		this.parent = parent;
		this.name = name;
	}

	/**
	 * Returns parent entity metamodel.
	 * This method returns <code>null</code>, when this instance is declared as 
	 * owned child entity property.
	 * 
	 * @return The parent entity metamodel.
	 */
	public Metamodel<?> getParent() {
		return parent;
	}

	/**
	 * Returns entity metamodel name.
	 * This method returns empty string, when this instance is declared as owned 
	 * child entity property.
	 * 
	 * @return The entity metamodel name.
	 */
	public String getName() {
		return name;
	}
	
	/**
	 * Returns this entity metamodel instance is root entity or not.
	 * 
	 * @return This entity metamodel instance is root entity or not.
	 */
	public boolean isRoot() {
		return (getParent() == null);
	}
	
	/**
	 * Returns the entity kind this metamodel represents.
	 * 
	 * @return The entity kind this metamodel represents.
	 */
	public Class<E> getKind() {
		return kind;
	}
	
	/**
	 * Returns the root entity kind of this metmaodel entity instance.
	 * 
	 * @return The root entity kind of this metmaodel entity instance.
	 */
	public Class<?> getRootKind() {
		if (isRoot()) {
			return getKind();
		} else {
			return getParent().getRootKind();
		}
	}
	
}
